/*
 *  $Id: mpir.h,v 1.8 1994/01/28 17:42:44 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

/* include file for the MPIR implementation of MPI */

#include <stdio.h>

#ifndef _MPIR_INCLUDE
#define _MPIR_INCLUDE

#include "mpid.h"

/*****************************************************************************
*                           MPIR DATA STRUCTURES                             *
*****************************************************************************/

/* Should switch these to MPIRMALLOC, MPIRFREE, MPIRNEW */
#ifndef MALLOC
#define MALLOC(a)    malloc((unsigned)(a))
#define FREE(a)      free((char *)(a))
#define NEW(a)    (a *)MALLOC(sizeof(a))
#endif

typedef struct _MPIR_DATATYPE {
    MPIR_NODETYPE dte_type;	/* type of datatype element this is */
    MPIR_BOOL committed;	/* whether committed or not */
    int is_contig;              /* whether entirely contiguous */
    int extent;			/* extent of this datatype */
    int use_count;		/* nodes depending on this node */
    struct _MPIR_DATATYPE *old_type; /* type this type is built out of, if 1 */
    int count;			/* replication count */
    int stride;			/* stride, for VECTOR and HVECTOR types */
    int blocklen;		/* blocklen, for VECTOR and HVECTOR types */
    int *indices;		/* array of indices, for (H)INDEXED */
    int *blocklens;		/* array of blocklens for (H)INDEXED */
    struct _MPIR_DATATYPE **old_types;  /* array of types, for STRUCT */
} MPIR_DATATYPE;

typedef enum {
    MPIR_SEND,
    MPIR_RECV
} MPIR_OPTYPE;

/* MPIR_COMMON is a subset of the handle structures that contain common
   elements.  MAKE SURE THAT ANY CHANGES IN THE HANDLES PRESERVES
   THIS ORDER! */
typedef struct {
    MPIR_OPTYPE handle_type;
    int         partner;        /* source or destination, depending on type */
    int         tag;
    int         contextid;
    MPIR_COMMUNICATOR *comm;
    MPIR_BOOL   completed;
    MPIR_DATATYPE *datatype;
    MPI_Persistence persistence; /* persistence: PERSISTENT or EPHEMERAL */

    void *bufadd;		
    int buflen;			
    int count;			

    char *bufpos;		
    int  transfer_count;	
    int  totallen;		
    } MPIR_COMMON;

typedef struct {
    MPIR_OPTYPE handle_type;	/* send or receive */
    int         dest;		/* destination process for message */
    int         tag;		/* tag */
    int         contextid;	/* context id */
    MPIR_COMMUNICATOR *comm;	/* communicator */
    MPIR_BOOL   completed;	/* whether operation is completed or not */
    MPIR_DATATYPE *datatype;	/* basic or derived datatype */
    MPI_Persistence persistence; /* persistence: PERSISTENT or EPHEMERAL */
    void *bufadd;		/* address of buffer */
    int buflen;			/* length of buffer at bufadd */
    int count;			/* requested count */
    char *bufpos;		/* position of next byte to be transferred */
    int  transfer_count;	/* count of bytes transferred to/from device */
    int  totallen;		/* length in local bytes of message */

    MPI_Mode mode;		/* mode: STANDARD, READY, SYNC */
    MPID_SHANDLE dev_shandle;	/* device's version of send handle */
} MPIR_SHANDLE;

typedef struct {
    MPIR_OPTYPE handle_type;	/* send or receive */
    int         source;		/* source process message */
    int 	tag;		/* tag */
    int         contextid;	/* context id */
    MPIR_COMMUNICATOR *comm;	/* communicator */
    MPIR_BOOL   completed;	/* Whether this receive has been completed */
    MPIR_DATATYPE *datatype;	/* basic or derived datatype */
    MPI_Persistence persistence; /* persistence: PERSISTENT or EPHEMERAL */
    void *bufadd;		/* address of buffer */
    int buflen;			/* length of buffer at bufadd */
    int count;			/* requested count */
    char *bufpos;		/* position of next byte to be transferred */
    int  transfer_count;	/* count of bytes transferred to/from device */
    int  totallen;      	/* length in local bytes of message */

    void *p;			/* Pointer to unexpected data */
    int  len;			/* Length of this data ? */
    int  actcount;		/* number of items actually read */
    MPID_RHANDLE dev_rhandle;	/* device's version of recv handle */
} MPIR_RHANDLE;

#define MPIR_HANDLES_DEFINED

typedef union {
    MPIR_OPTYPE type;		/* send or receive */
    MPIR_COMMON  chandle;       /* common fields */
    MPIR_SHANDLE shandle;
    MPIR_RHANDLE rhandle;
} MPIR_HANDLE;

typedef struct _MPIR_Intercomm_request {
  int tag;                /* Unique tag */
  MPIR_COMMUNICATOR *local_comm;
  int                local_leader;
  int                local_size;
  MPIR_COMMUNICATOR *peer_comm; 
  int                remote_leader;
  int                remote_size;
  MPIR_CONTEXT       recv_context;
  MPIR_CONTEXT       send_context;
  MPIR_HANDLE       *req[6];
} MPIR_Intercomm_request;

typedef enum {
    MPIR_QSHANDLE,
    MPIR_QRHANDLE
} MPIR_QEL_TYPE;

typedef struct _MPIR_QEL {	/* queue elements */
    void *ptr;			/* queues can contain anything  */
    MPIR_QEL_TYPE qel_type;	/* but we may need to know what */
    struct _MPIR_QEL *prev;	/* previous queue element */
    struct _MPIR_QEL *next;	/* next queue element */
} MPIR_QEL;

typedef struct {		/* header for queues of things like handles */
    MPIR_QEL *first;		/* first queue element */
    MPIR_QEL *last;		/* last queue element */
    int      currlen;		/* current length */
    int      maxlen;		/* maximum since beginning of run */
} MPIR_QHDR;

typedef struct _MPIR_FDTEL {	/* flat datatype elements */
    MPIR_NODETYPE      type;	/* one of the elementary data types */
    int                disp;	/* displacement */
    struct _MPIR_FDTEL *next;	/* pointer to next element */
} MPIR_FDTEL;

#define MPIR_MALLOC(size)            malloc(size)
#define MPIR_dump_datatype(datatype) MPIR_dump_dte(datatype,0)

/*****************************************************************************
*                                GLOBAL DATA                                 *
*****************************************************************************/

/* memory management for fixed-size blocks */
void *MPIR_shandles;		/* sbcnst MPIR_SHANDLES */
void *MPIR_rhandles;		/* sbcnst MPIR_RHANDLES */
void *MPIR_dtes;		/* sbcnst datatype elements */
void *MPIR_qels;		/* sbcnst queue elements */
void *MPIR_fdtels;		/* sbcnst flat datatype elements */
void *MPIR_hbts;        /* sbcnst height balanced tree roots for cacheing */
void *MPIR_hbt_els;     /* sbcnst height balanced tree nodes for cacheing */

/* pre-assigned datatypes */
MPIR_DATATYPE *MPIR_int_dte;
MPIR_DATATYPE *MPIR_float_dte;
MPIR_DATATYPE *MPIR_double_dte;
MPIR_DATATYPE *MPIR_long_dte;
MPIR_DATATYPE *MPIR_complex_dte; /* note not an MPI type, only MPIR */
MPIR_DATATYPE *MPIR_short_dte;
MPIR_DATATYPE *MPIR_char_dte;
MPIR_DATATYPE *MPIR_byte_dte;
MPIR_DATATYPE *MPIR_uchar_dte;
MPIR_DATATYPE *MPIR_ushort_dte;
MPIR_DATATYPE *MPIR_ulong_dte;
MPIR_DATATYPE *MPIR_uint_dte;

/* queues */
MPIR_QHDR posted_sends;
MPIR_QHDR posted_recvs;
MPIR_QHDR unexpected_recvs;

/* MPIR routines */
extern MPIR_GROUP *MPIR_CreateGroup();

/* MPIR process id (from device) */
int MPIR_tid;

#endif
